/*******************************************************************************
 client.js: main client file (the lower level, under viewer.js)

 $Author: dmc $
 $Revision: 2607 $
 $Date: 2019-06-12 17:15:12 -0300 (Wed, 12 Jun 2019) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/client.js $
*******************************************************************************/

/*
  - handles connection to server
  - maintains a arrays of frame, face, string, and shot objects
  - call init_client() to initialize
  - implement update() to be called to redraw screen when things change
*/

// *** globals accessible by client ***

var g_socket;

var g_frames = {}; // ClientFrame
var g_faces = {}; // ClientFace
var g_strings = {}; // ClientString
var g_devices = {}; // ClientDevice
var g_users = {}; // ClientUser
var g_config = {}; // system configuration options
var g_events = {}; // events, matches, stages

var g_archive = {}; // array of inactive strings in order (no shots)

rstat_log = []; // messages received from rstat sync

heartbeats = []; // store heartbeats received while waiting for all data

// clears on broadcast
function clear_everything() {
	g_frames = {};
	g_faces = {};
	g_strings = {};
	g_devices = {};
	g_users = {};
	g_config = {};
	g_archive = {};
	g_events = {};
}

var g_string_view = null; // currently viewing inactive string

var g_connected = false; // boolean whether server is OK

var loaded_images = {}; // images indexed by src

var cur_sequence_id = 0;
var cur_server_ts = 0;
var cur_interval = 1500; // gets updated from msg.interval
var cur_ts_local_error = 0;

var init_data_received = false;

var recent_frame = -1;

var remote_logging = false;

// options

var fix_errors = true; // set true to enable auto fixing database errors

// new empty message for server
function new_crud() { return { frames: {}, faces: {}, strings: {}, devices: {}, users: {}, config: {}, events: {} } }

function config_crud(delta) {
	var msg = { config: delta };
	log("config_crud()", msg);
	g_socket.emit("crud", msg);
}

function frame_id_crud(frame_id, delta) {
	var msg = { frames: {} }; msg.frames[frame_id] = delta;
	log("frame_id_crud()", msg);
	g_socket.emit("crud", msg);
}

function log_on() {
	config_crud({enable_logging: true});
}

function log_off() {
	config_crud({enable_logging: false});
}

// simulate lots of shots
// meant to be used from web console
var sim_timer = null;
function sim(n, interval) {
	clearInterval(sim_timer);
	if (interval) {
		sim_timer = setInterval(function() {
			sim_actual(n);
		}, interval * 1000);
		return;
	}
	sim_actual(n);	
}

function sim_actual(n) {
	for (i = 0; i < n; i++) {
		var x = 1.2*254 * rndn();
		var y = 1.5*254 * rndn();
//		var v = 400 + 300 * rnd();
//		var theta = rndu() * 0.05;
//		var phi = rndu() * 0.05;
//		g_socket.emit("frame_simulate_shot", cur_string().frame().id(), x, y, v, theta, phi, 20);
		client_simulate_shot_xy(x,y);
	}
}

function client_simulate_shot(x, y, v, theta, phi, temp) {
	g_socket.emit("frame_simulate_shot", cur_string().frame().id(), x, y, v, theta, phi, temp);
}

function client_simulate_shot_xy(x, y) {
	var v = (1400 + rndn() * 20) / 3.28; // fps->m/s
	var theta = (3 + rndn()*4)/1000; // mils->rad
	var phi = (-10 + rndn()*6)/1000; // mils->rad
	var temp = 23 + rndn()*1; // deg-C

	g_socket.emit("frame_simulate_shot", cur_string().frame().id(), x, y, v, theta, phi, temp);
}

// new way for client to request all data
function request_data() {
	init_data_received = false; // ignore future heartbeats
	heartbeats = []; // reset saved heartbeats
	request_data_actual();
}

// client ask server for full data set
function request_data_actual() {
	var xhr = new XMLHttpRequest();
	// add random parameter to prevent caching
	xhr.open("GET", "/data?" + new Date().getTime()); // async
	xhr.onload = function (e) {
	  if (xhr.readyState === 4) {
	    if (xhr.status === 200) {
	      console.log("GET /data received, size=",xhr.responseText.length);
	      var obj = JSON.parse(xhr.responseText);
		  recv_server_data(obj, true);
		  
		  // process any heartbeats received in the meantime
		  for (var seq = obj.seq + 1; heartbeats[seq]; seq++) {
		  	console.log("processing heartbeat " + seq);
		  	recv_server_data(heartbeats[seq], false);
		  	heartbeats[seq] = null;
		  }
	    } else {
	      console.error(xhr.statusText);
	    }
	  }
	};
	xhr.onerror = function (e) {
	  console.error(xhr.statusText);
	};
	xhr.send();
}

// DMC 26-Feb-2017:
function request_get_archive(user_id, event_id, final, recent, only_framename) {

	console.log("requesting GET /archive");
	
	var xhr = new XMLHttpRequest();
	// add random parameter to prevent caching
	var request;
	xhr.open("GET", request = "/archive?" + new Date().getTime()
	+'&user_id='+encodeURIComponent(user_id)
	+'&event_id='+encodeURIComponent(event_id)
	+'&final='+encodeURIComponent(final)
	+'&recent='+encodeURIComponent(recent)
	+'&only_framename='+encodeURIComponent(only_framename)
	, true);

	xhr.onload = function (e) {
	  if (xhr.readyState === 4) {
	    if (xhr.status === 200) {
	      console.log("GET /archive received, size=",xhr.responseText.length,"type=",typeof xhr.responseText);
	      var obj = undefined;
	      try {
		      obj = JSON.parse(xhr.responseText);
		      // console.log("parsed=",obj);
	      } catch (e) {
		      console.log("/archive parse error",e);
	      }
	      try {
			  receive_archive(obj); // GET /archive returns array of archive items
	      } catch (e) {
		      console.log("receive_archive() error",e);
	      }
	    } else {
	      console.error(xhr.statusText);
	    }
	  }
	};
	xhr.onerror = function (e) {
	  console.error(xhr.statusText);
	};
	xhr.send();
}

function receive_archive(msg) { // old-style; expects msg to be archive data
	g_archive = msg;
	update();
};


// *** main entry function called by init_targetviewer ***
function init_client() {
	log("init_client()");
	
	g_socket = io(); // connection

	// get all data on connect
	g_socket.on("connect", function() {
		g_socket.emit("reset_time_load", Date.now());
		request_data();
	});

	// respond to server request for client to get new data (broadcast does this)
	g_socket.on("request_data", function() {
		request_data();
	});

	// received heartbeat from server
	g_socket.on("delta", function(msg) {
		recv_server_data(msg, false);
//		console.log("DMC",g_config.is_client_return_ping);
		if (g_config.is_client_return_ping) {
			var return_msg = { ts: msg.ts, seg:msg.seq };
			g_socket.emit("client_return_ping",return_msg);
		}
	});
	
	g_socket.on("archive", receive_archive);
	
	g_socket.on("string", function(string) {
		g_string_view = new ClientString(string);
		select_string(g_string_view.id());
		if (toolbar.active_panel != toolbar_review) {
			if (toolbar.back()) toolbar.set_panel(toolbar_review);
		}
		update();
	});
	
	g_socket.on("download_file", function(file) {
		log("downloading " + file);
		window.location.assign(file);
		setTimeout(function() {
			window.location.reload();
		}, 1000);
	});
	
	g_socket.on("update", function(msg) {
		update();
	});
	
	g_socket.on("set_remote_log", function(val) {
		remote_logging = val;
	});

	g_socket.on("log", function(msg) {
		log("server-log:", msg);
	});
		
	g_socket.on("rstat_log", function(msg, string, shooter_i) {
		log("got rstat log message: ", msg);
		rstat_log.push({text: msg, string: string, shooter_i: shooter_i});
		toolbar.update();
	});
	
	g_socket.on("client_refresh", function() {
		window.location.reload();
	});
	
	g_socket.on("results", function(results) {
//		console.log("client.js RESULTS",results);
		
		var scores_tr = [];
		var scores_f = [];
		
		for (i in results) {
			
			var score = {name: user_display_name(i), score: results[i]};
			
			if (g_users[i].data.fclass_target) scores_f.push(score);
			else scores_tr.push(score);			
			
		}
		
		scores_tr.sort(function(a,b) {
		    if (a.score > b.score) return -1;
		    if (a.score < b.score) return 1;
		    return 0;
		});

		scores_f.sort(function(a,b) {
		    if (a.score > b.score) return -1;
		    if (a.score < b.score) return 1;
		    return 0;
		});
		
		if (scores_tr.length > 0) {
			var str = "TR:\n";
			for (i in scores_tr) {
				str += scores_tr[i].score.toFixed(3) + " " + scores_tr[i].name + "\n";
			}
		}

		if (scores_f.length > 0) {
			str += "\nF:\n";
			for (i in scores_f) {
				str += scores_f[i].score.toFixed(3) + " " + scores_f[i].name + "\n";
			}
		}
		
		alert(str);

	});

}

function show_spinner () {
    var mainTarget = $('#divMainTarget');
    var spinner = $('#debug_spinner');
    if (mainTarget !== undefined && spinner !== undefined) {
        var left = (mainTarget.width() / 2) - (spinner.width() / 2) + 5;
        var top = (mainTarget.height() / 2) - (spinner.height() / 2) + 5;
        $('#debug_spinner').css('left', left).css('top', top);
        $('#debug_spinner').fadeIn(900);        
    }
}

function hide_spinner () {
    var spinner = $('#debug_spinner');
    if (!spinner.is(":visible")) return false;
    spinner.fadeOut();
    return true;
}

// general purpose data received from server
function recv_server_data(msg, complete) {
	if (complete) log(msg.seq, "received all data message", msg);
	else log(msg.seq, "received heartbeat", msg);
    
    if (link_disconnect_timer !== undefined) {
        hide_spinner();
        clearTimeout(link_disconnect_timer);
        link_disconnect_timer = undefined;
    }
        
	// if have not received initial data yet, ignore heartbeats
	if (!complete && !init_data_received) {
		log("storing heartbeat in queue");
		heartbeats[msg.seq] = msg;
		return;
	}
	init_data_received = true;
	
	// generic error message after data is received
	$("#spanLoadingText").text("Data received from server...");

	// heartbeat sync error: request all data
	if (!complete && msg.seq != cur_sequence_id && msg.seq != (cur_sequence_id + 1)) {
		console.log("SYNC ERROR!!!!");
		request_data(); // this will only happen once since init_data_received is now false
		return;
	}

	cur_sequence_id = msg.seq;
	if (msg.ts) cur_server_ts = msg.ts;
	if (msg.interval) cur_interval = msg.interval;
	cur_ts_local_error = Date.now() - cur_server_ts;
	
	// update display only if something changes
	var need_update = complete; // always update after initial
	
	if (complete) clear_everything(); // all data: reset arrays
	

	function apply_msg_group(ctor/*ClientFrame, etc*/, g_var/*g_frames*/, msg_section/*msg.frame, msg.users etc*/) {
		for (var i in msg_section) {
			// need_update = true; // set only if update_object() tells us it has changed something
			var msg_itm = msg_section[i];
			var delete_flag = test_object_delete(msg_itm);
			if (!delete_flag && ! g_var[i]) {
				g_var[i] = new ctor(msg_itm);
				need_update = true;
			} else if (delete_flag) {
				delete g_var[i];
				need_update = true;
			} else {
				var uo_opts = {record_dirty: true, trace_dirty: false };
				update_object(g_var[i].data, msg_itm, uo_opts);
				need_update = need_update || uo_opts._dirty; // only set if data changes
				//if (uo_opts._dirty) console.log("DIRTY",msg_itm);
			}
		}
	}

	apply_msg_group(ClientFrame,g_frames, msg.frames);
	apply_msg_group(ClientFace,g_faces, msg.faces);
	// special additional integrity check for strings:
	var strings_to_process = undefined;
	for (var i in msg.strings) {
		// if we don't recognize the string and it doesn't have a frame or face, ignore it
		if (!g_strings[i] && (!msg.strings[i].frame_id || !msg.strings[i].face_id)) continue;
		if (undefined == i) continue; // ignore key=='undefined'
		strings_to_process = strings_to_process || {};
		strings_to_process[i]=msg.strings[i];
	}
//	apply_msg_group(ClientString,g_strings, strings_to_process);
	apply_msg_group(ClientString,g_strings, msg.strings);
	apply_msg_group(ClientDevice,g_devices, msg.devices);
	apply_msg_group(ClientUser,g_users, msg.users);

	// strings need more work than apply_msg_group does:
	if (toolbar_info && 'function' == typeof toolbar_info.checkStringForStrikeOutShots) {
		for (var i in msg.strings) {
            // let the info toolbar decide if some stuff is to do
            toolbar_info.checkStringForStrikeOutShots(msg.strings[i], g_strings[i]);
        }
    }

	// update any configuration options
	if (msg.config) {
		update_object(g_config, msg.config);
		need_update = true;
		// check if we should start pinging the server...
		doPing();
	}
	
	// update any configuration options
	if (msg.events) {
		update_object(g_events, msg.events);
		need_update = true;
	}

	if (!check_everything()) return;
    
	// server is connected
	if (link_reset()) need_update = true;

	// let the application react to new data
	if (typeof(data_received) != "undefined") data_received(msg, complete);

	if (need_update) {
		log("data from server: ", msg);
		update();
	}
	
}

function fix_errors_now() {
	fix_errors = true;
	check_everything();
}

// check and crash predictably each time data is received
function check_everything() {
	
	function fail(msg) { // we cannot afford to fail...
//		clear_everything();
//		init_data_received = false;
		alert(msg + "\n\n*******************\nPlease report this to SMT or your dealer" );
	}
	
	// there must be at least one frame and face
	if (numberof(g_frames) == 0 || numberof(g_faces) == 0) {
		fail("There must be at least one frame and face!");
return true; // we cannot afford to fail...
		return false;
	}
	
	// make sure each frame has a current string
	// if not, request a new string be created
	var error = 0;
	for (var i in g_frames) {
		if (!g_frames[i].string()) {
			//if (fix_errors) ClientString.crud({ frame_id: i }); // this doesn't work
			log(g_frames[i]);
			error++;
		}
	}
	if (error > 0) {
		fail(error + " frames do not have a valid current string!");
return true; // we cannot afford to fail...
		return false;
	}

	// check to make sure each string has a valid face
	// if not select one
	var error = 0;
	for (var i in g_strings) {
		if (!g_strings[i].face()) {
console.log("check_everything() fix string/face",i,g_strings[i]);
			if (fix_errors) g_strings[i].crud({ face_id: first_key(g_faces) });
			log(g_strings[i]);
			error++;
		}
	}
	if (error > 0) {
		fail(error + " strings do not have a valid face!");
return true; // we cannot afford to fail...
		return false;
	}

	// check to make sure each string has a valid frame
	var error = 0;
	for (var i in g_strings) {
		if (!g_strings[i].frame()) {
			log(g_strings[i]);
			error++;
		}
	}
	if (error > 0) {
		fail(error + " strings do not have a valid frame!");
return true; // we cannot afford to fail...
		return false;
	}
	
	return true;
}


// for server connection timeout
var link_timer;
var link_disconnect_timer;


// timer went off, server must be down
function link_expired() {
    show_spinner();
	var timeoutRedX = g_config.option_timeout_red_x || 10000;
	if (timeoutRedX < 1 || timeoutRedX > 20000) timeoutSpinner = 10000;
    link_disconnect_timer = setTimeout(function() {
        hide_spinner();
        g_connected = false; // trigger red-X
        update();
    }, timeoutRedX); // DMC 9 April 2017
}

// call when message is received by the server to update connection status
// returns whether status has changed
// provide time in milliseconds expected to receive next message within
function link_reset() {
	var need_update = false;
	if (g_connected == false) need_update = true;
	g_connected = true;
	clearTimeout(link_timer); // reset existing timer
	var timeoutSpinner = g_config.option_timeout_spinner || 500;
	if (/*timeoutSpinner < 1 ||*/ timeoutSpinner > 10000) timeoutSpinner = 500;
//console.log("dmc: timeoutSpinner",timeoutSpinner,cur_interval);
	link_timer = setTimeout(link_expired, cur_interval + timeoutSpinner); // heartbeat expected time plus timeout value in ms
	return need_update;
}


// efficiently handle window resizing
var RESIZE_TIMEOUT = 100; // timeout in milliseconds
var resize_timer; // timer variable
$(window).resize(function() {
	clearTimeout(resize_timer);
	resize_timer = setTimeout(update, RESIZE_TIMEOUT);
});

// cookie monsters

function set_cookie(name, value) {
	document.cookie = name + "=" + value + ";";
}

function get_cookie(name) {
	var ca = document.cookie.split(";");
	for (var i = 0; i < ca.length; i++) {
		var c = ca[i];
		while (c.charAt(0) == " ") c = c.substring(1);
		if (c.indexOf(name + "=") == 0) return c.substring((name + "=").length, c.length);
	}
	return "";
}

function clear_cookie(name) {
	document.cookie = name + "=; expires=Thu, 01 Jan 1970 00:00:00 UTC";
}

// replace touchstart with mousedown unless allow_touchstart is set on event
// this allows mousedown event to be used for immediate touch response
// also disables panning/scrolling

// New Chrome update end of Feb 2017 ruins this, so comment it out for now:
/*
$(document).bind("touchstart", function(e) {
	if (e.target.className.substring(0, 3) == "ui-") return; // not for jquery-ui elements
	if (e.originalEvent.allow_touchstart) return;
	e.preventDefault();
//console.log("DMC--touchstart handler - sending mousedown");
	$(e.target).trigger("mousedown");
});
*/

// ignore touchmove to disable scrolling unless allow_touchmove is set on event
// set this to allow scrolling only in specific cases
$(document).bind("touchmove", function(e) {
	if (e.originalEvent.allow_touchmove) return;
	e.preventDefault();
});

window.onerror = function(msg, url, line, col, error) {
   // Note that col & error are new to the HTML 5 spec and may not be 
   // supported in every browser.  It worked for me in Chrome.
   var extra = !col ? '' : '\ncolumn: ' + col;
   extra += !error ? '' : '\nerror: ' + error;

   // You can view the information in an alert to see things working like this:
   alert("Error: " + msg + "\nurl: " + url + "\nline: " + line + extra);

   // TODO: Report this error via ajax so you can keep track
   //       of what pages have JS issues

   var suppressErrorAlert = true;
   // If you return true, then error alerts (like in older versions of 
   // Internet Explorer) will be suppressed.
   return suppressErrorAlert;
};

var alreadyPinging = false;

function doPing(force) {
    if (g_config.client_ping && (!alreadyPinging || force)) {
        alreadyPinging = true;
        var ts = Date.now();
        $.get('/ping', {}, function(result) {
            document.title = 'SMT ' + result + ' (' + (Date.now() - ts) + 'ms)';
            setTimeout(function() {
                doPing(true);
            }, 1000);
        });
    } else if (!g_config.client_ping) {
        document.title = 'Silver Mountain Targets';
        alreadyPinging = false;
    }
}
